# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Test cleaning the HTML input."""

from pprint import pprint

import pytest

from open_web_calendar.clean_html import clean_html, remove_html


@pytest.mark.parametrize(
    ("html", "spec", "expected_output"),
    [
        ("<script>js</script>a", {"clean_html_scripts": True}, "a"),
        ("<malicious>", {"clean_html_remove_unknown_tags": True}, ""),
    ],
)
def test_clean_html_from_spec(html, spec, expected_output):
    """Specify how to clean the HTML content."""
    output = clean_html(html, spec=spec).replace("\n", "")
    assert output == expected_output


def test_remove_html():
    """Remove all HTML tags - we do not expect them in the summary."""
    assert remove_html("<asd>aaa</asd>bbb") == "aaabbb"
    assert remove_html("><<<>") == ">"


def test_nice_id_for_event(client, calendar_urls):
    """Make sure that we have nice IDs without HTML in them."""
    response = client.get(
        f"/calendar.events.json?url={calendar_urls['one-event']}&timezone=Europe/London&from=2019-03-04&to=2019-03-05"
    )
    event = response.json[0]
    pprint(event)
    assert " " not in event["id"]
    assert ":" not in event["id"]



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

import datetime

import pytest
from pytz import timezone, utc

from open_web_calendar.app import ConvertToDhtmlx

berlin = timezone("Europe/Berlin")
eastern = timezone("US/Eastern")

UTC = "UTC"
BERLIN = "Europe/Berlin"
EASTERN = "US/Eastern"


@pytest.mark.parametrize(
    ("date", "tz", "expected"),
    [
        # test date object without time zone
        (datetime.date(2019, 3, 23), UTC, "2019-03-23 00:00"),
        (datetime.date(2019, 3, 23), BERLIN, "2019-03-23 00:00"),
        (datetime.date(2019, 3, 23), EASTERN, "2019-03-23 00:00"),
        # test datetime object without time zone
        (datetime.datetime(2019, 6, 2, 20), UTC, "2019-06-02 20:00"),
        (datetime.datetime(2019, 4, 20, 10, 10), BERLIN, "2019-04-20 10:10"),
        (datetime.datetime(2019, 12, 1, 0, 13, 23), EASTERN, "2019-12-01 00:13"),
        # test datetime object with time zone
        (berlin.localize(datetime.datetime(2019, 6, 2, 20)), UTC, "2019-06-02 18:00"),
        (
            berlin.localize(datetime.datetime(2019, 4, 20, 10, 10)),
            BERLIN,
            "2019-04-20 10:10",
        ),
        (
            berlin.localize(datetime.datetime(2019, 12, 1, 0, 13, 23)),
            EASTERN,
            "2019-11-30 18:13",
        ),
        (utc.localize(datetime.datetime(2019, 6, 2, 20)), UTC, "2019-06-02 20:00"),
        (
            utc.localize(datetime.datetime(2019, 4, 20, 10, 10)),
            BERLIN,
            "2019-04-20 12:10",
        ),
        (
            utc.localize(datetime.datetime(2019, 12, 1, 0, 13, 23)),
            EASTERN,
            "2019-11-30 19:13",
        ),
    ],
)
def test_date_to_string_conversion(date, tz, expected):
    """Convert dates and datetime objects for the events.json"""
    string = ConvertToDhtmlx({"timezone": tz}).date_to_string(date)
    print(date)
    assert string == expected



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Browser fixture setup and teardown

see https://behave.readthedocs.io/en/stable/practical_tips.html#selenium-example
"""

import copy
import http.server
import multiprocessing
import random
import socketserver
import subprocess
import sys
import threading
import time
from pathlib import Path

import requests
from behave import fixture, use_fixture
from selenium import webdriver
from selenium.webdriver import Firefox, FirefoxOptions
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.firefox.service import Service
from werkzeug import run_simple

HERE = Path(__file__).parent.absolute()
sys.path.append(HERE / "..")
from open_web_calendar.app import DEFAULT_SPECIFICATION, app  # noqa: E402

CALENDAR_FOLDER = HERE / "calendars"
# timeout in seconds
WAIT = 10


def locate_command(command: str):
    """Locate a command on the command line or return ''."""
    code, output = subprocess.getstatusoutput(f"which {command}")  # noqa: S605
    if code == 0:
        return output.strip()
    return ""


@fixture
def browser_firefox(context):
    # run firefox in headless mode
    # see https://stackoverflow.com/a/47642457/1320237
    opts = FirefoxOptions()
    opts.add_argument("--headless")
    # specify firefox executible and gecko drivers
    # see https://stackoverflow.com/a/76852633
    # see https://stackoverflow.com/a/71766991/1320237
    # specify the path to your geckodriver
    geckodriver_path = Path("/snap/bin/geckodriver")
    # Set the language for the tests
    opts.set_preference("intl.accept_languages", "en-US, en")
    # construct the arguments
    kw = {"options": opts}
    if geckodriver_path.exists():
        kw["service"] = Service(executable_path=geckodriver_path)
    browser = Firefox(**kw)
    context.browser = browser
    browser.set_page_load_timeout(10)
    yield context.browser
    # -- CLEANUP-FIXTURE PART:
    context.browser.quit()


@fixture
def browser_chrome(context):
    # see https://behave.readthedocs.io/en/stable/tutorial.html#environmental-controls
    # for an example
    # lots of options come from https://stackoverflow.com/a/52340526/1320237
    options = webdriver.ChromeOptions()
    options.add_argument(
        "--headless"
    )  # from https://stackoverflow.com/q/56637973/1320237
    options.add_argument(
        "start-maximized"
    )  # https://stackoverflow.com/a/26283818/1689770
    options.add_argument(
        "enable-automation"
    )  # https://stackoverflow.com/a/43840128/1689770
    options.add_argument("--no-sandbox")  # https://stackoverflow.com/a/50725918/1689770
    options.add_argument(
        "--disable-dev-shm-usage"
    )  # https://stackoverflow.com/a/50725918/1689770
    options.add_argument(
        "--disable-browser-side-navigation"
    )  # https://stackoverflow.com/a/49123152/1689770
    options.add_argument(
        "--disable-gpu"
    )  # https://stackoverflow.com/questions/51959986/how-to-solve-selenium-chromedriver-timed-out-receiving-message-from-renderer-exc

    # executable_path from https://stackoverflow.com/a/76550727/1320237
    path = locate_command("chromium.chromedriver") or locate_command("chromedriver")
    service = Service(executable_path=path)
    context.browser = webdriver.Chrome(service=service, options=options)
    yield context.browser
    context.browser.quit()


@fixture
def browser_safari(context):
    context.browser = webdriver.Safari()
    yield context.browser
    context.browser.quit()


browsers = {
    "firefox": browser_firefox,
    "chrome": browser_chrome,
    "safari": browser_safari,
}


@fixture
def set_window_size(context):
    """Set the window size of the browser."""
    # see https://stackoverflow.com/a/55878622/1320237
    if "window" in context.config.userdata:
        width, height = context.config.userdata["window"].split("x")
        context.browser.set_window_size(int(width), int(height))


def get_free_port(start=10000, end=60000):
    """Return a free port number."""
    return random.randint(start, end)  # noqa: S311


@fixture
def app_server(context):
    """Start the flask app in a server."""
    app_port = get_free_port()
    # from https://werkzeug.palletsprojects.com/en/2.1.x/serving/#shutting-down-the-server
    # see also https://stackoverflow.com/questions/72824420/how-to-shutdown-flask-server
    p = multiprocessing.Process(target=run_simple, args=("localhost", app_port, app))
    p.start()
    context.index_page = f"http://localhost:{app_port}/"
    wait_for_http_server(context.index_page, on_error=p.terminate)
    yield
    p.terminate()


def wait_for_http_server(url, on_error=lambda: None):
    """Make sure the HTTP server is up and running."""
    print(f"HTTP SERVER: Waiting for {url} to start ... ", end="")
    timeout = time.time() + WAIT
    ty = err = tb = None
    while time.time() < timeout:
        try:
            requests.get(url, timeout=WAIT)
            ty = err = tb = None
            break
        except:
            ty, err, tb = sys.exc_info()
            time.sleep(0.01)
    if err is not None:
        print("FAIL")
        on_error()
        raise err.with_traceback(tb)
    print("OK")


@fixture
def calendars_server(context):
    """Serve the calendar files so they can be requested.

    see https://stackoverflow.com/a/52531444/1320237
    """
    # reuse address
    # see https://zaiste.net/posts/python_simplehttpserver_not_closing_port/
    socketserver.TCPServer.allow_reuse_address = True
    port = 8001
    host = "localhost"
    context.calendars_url = f"http://{host}:{port}/"

    class Handler(http.server.SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, directory=CALENDAR_FOLDER, **kwargs)

    try:
        httpd = socketserver.TCPServer((host, port), Handler)
    except OSError as e:
        if e.errno == 98:
            # OSError: [Errno 98] Address already in use
            wait_for_http_server(context.calendars_url)
            yield
            return
        else:
            raise
    # from https://werkzeug.palletsprojects.com/en/2.1.x/serving/#shutting-down-the-server
    # see also https://stackoverflow.com/questions/72824420/how-to-shutdown-flask-server
    t = threading.Thread(target=httpd.serve_forever)
    t.start()

    def final():
        httpd.server_close()
        httpd.shutdown()

    wait_for_http_server(context.calendars_url, on_error=final)
    yield
    final()


def before_all(context):
    browser = browsers[context.config.userdata["browser"]]
    use_fixture(browser, context)
    use_fixture(set_window_size, context)
    use_fixture(app_server, context)
    use_fixture(calendars_server, context)


# Set the default timezone of the browser
# If we would like to set another timezone in the tests, we can write:
#    Given we set the "teimzone" parameter to "Asia/Singapore"
DEFAULT_SPECIFICATION.update(
    {
        "url": [],
        "timezone": "Europe/Moscow",
        "date": "1605-11-05",
    }
)


def before_scenario(context, scenario):
    """Reset the calendar for each scenario.

    Empty url and set the timezone and other parameters.
    """
    context.specification = copy.deepcopy(DEFAULT_SPECIFICATION)



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

from .app import app, main
from .version import __version__, __version_tuple__, version, version_tuple

__all__ = [
    "main",
    "app",
    "__version__",
    "version",
    "__version_tuple__",
    "version_tuple",
]



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""This tests that events receive the right styling.

See https://github.com/niccokunzmann/open-web-calendar/issues/305
"""

from __future__ import annotations

from typing import NamedTuple

import pytest

from open_web_calendar.app import ConvertToDhtmlx


class CATEGORY(NamedTuple):
    cats: list[str]


@pytest.mark.parametrize(
    ("event", "expected_classes"),
    [
        ({"UID": "test123"}, ["UID-test123"]),
        ({"CATEGORIES": CATEGORY(["CAT1"])}, ["CATEGORY-CAT1"]),
        ({"TRANSP": "OPAQUE"}, ["TRANSP-OPAQUE"]),
        ({"TRANSP": "TRANSPARENT"}, ["TRANSP-TRANSPARENT"]),
        ({"CLASS": "x-custom-token"}, ["CLASS-x-custom-token", "CLASS-PRIVATE"]),
        ({"CLASS": "PRIVATE"}, ["CLASS-PRIVATE"]),
        ({"CLASS": "CONFIDENTIAL"}, ["CLASS-CONFIDENTIAL"]),
        ({"CLASS": "PUBLIC"}, ["CLASS-PUBLIC"]),
        ({"PRIORITY": 5}, ["PRIORITY-5"]),
    ],
)
def test_get_event_classes(event, expected_classes):
    """Check that event classes are correctly extracted."""
    dhx = ConvertToDhtmlx({"timezone": "Europe/London"})
    classes = dhx.get_event_classes(event)
    assert classes == expected_classes


@pytest.mark.parametrize(
    ("calendar", "start", "stop", "uid", "classes"),
    [
        (
            "event-with-categories",
            "2023-03-04",
            "2023-03-05",
            "UYDQSG9TH4DE0WM3QFL2J",
            [
                "event",
                "UID-UYDQSG9TH4DE0WM3QFL2J",
                "CATEGORY-APPOINTMENT",
                "CATEGORY-EDUCATION",
                "CALENDAR-INDEX-0",
            ],
        ),
        (
            "issue-135-free-busy",
            "2022-11-18",
            "2022-11-23",
            "83E6A2ED-D816-4E6F-94B5-251A101C6E36",
            [
                "UID-83E6A2ED-D816-4E6F-94B5-251A101C6E36",
                "STATUS-CONFIRMED",
                "TRANSP-OPAQUE",
            ],
        ),
        (
            "events-with-privacy",
            "2024-04-08",
            "2024-04-10",
            "e897cc75-e428-4696-baf1-d1e22eba7e2e-2",
            ["CLASS-PRIVATE", "CLASS-x-custom-token"],
        ),
    ],
)
def test_classes_of_events(client, calendar, start, stop, uid, classes, calendar_urls):
    """Make sure that events have certain classes in them."""
    response = client.get(
        f"/calendar.events.json?url={calendar_urls[calendar]}&timezone=UTC&timeshift=0&from={start}&to={stop}"
    )
    events = [event for event in response.json if event["uid"] == uid]
    assert len(events) >= 1
    for event in events:
        print(classes)
        print(event["css-classes"])
        assert set(classes) <= set(event["css-classes"])



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

import json
import re
from urllib.parse import urlencode, urljoin

from behave import given, then, when
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC  # noqa: N812
from selenium.webdriver.support.ui import Select, WebDriverWait

# default wait time in seconds
WAIT = 10


def specification_to_query(spec):
    """Convert the specification to a query string for the url."""
    result = []
    for k, vs in spec.items():
        for v in vs if isinstance(vs, list) else [vs]:
            result.append(urlencode({k: v}))
    return "&".join(result)


def get_url(context, url):
    """Replace getting the URL to mitigate a TimeoutException in Chrome."""
    print(
        f"Visiting {re.sub('^http://localhost:[0-9]+/', 'http://localhost:5000/', url)}"
    )
    for _i in range(20):
        try:
            return context.browser.get(url)
        except TimeoutException:  # noqa: PERF203
            pass
    raise  # noqa: PLE0704, RUF100


@given('we add the calendar "{calendar_name}"')
def step_impl(context, calendar_name):
    """Add a calendar to the url field.

    If there is no ending, .ics is added.
    """
    if "." not in calendar_name:
        calendar_name += ".ics"
    calendar_url = urljoin(context.calendars_url, calendar_name)
    context.specification["url"].append(calendar_url)


@given('we set the "{parameter_name}" parameter to {parameter_value}')
def step_impl(context, parameter_name, parameter_value):
    context.specification[parameter_name] = json.loads(parameter_value)


@when("we look at {date}")
def step_impl(context, date):
    context.specification["date"] = date
    context.browser.delete_all_cookies()
    url = (
        context.index_page
        + "calendar.html?"
        + specification_to_query(context.specification)
    )
    get_url(context, url)
    wait_for_calendar_to_load(context)


def wait_for_calendar_to_load(context):
    # wait until the loader has stopped spinning
    # see https://stackoverflow.com/a/53242626/1320237
    # see https://stackoverflow.com/a/26567563/1320237
    WebDriverWait(context.browser, WAIT).until(
        EC.presence_of_element_located(
            (By.XPATH, '//div[@id = "loader" and contains(@class, "hidden")]')
        )
    )


@then("we see 1 event")
@then("we see {count} events")
def step_impl(context, count=1):
    WebDriverWait(context.browser, WAIT).until(
        EC.presence_of_element_located((By.XPATH, "//div[contains(@class, 'event')]"))
    )
    events = context.browser.find_elements(
        By.XPATH, "//div[contains(@class, ' event ')]"
    )
    assert (
        len(events) == count
    ), f"Expected {count} events but {len(events)} were found: {events!r}"


@then('we see that event "{uid}" has the text "{text}"')
def step_impl(context, uid, text):
    events = context.browser.find_elements(
        By.XPATH, f"//div[contains(@event_id, {uid!r})]"
    )
    assert (
        len(events) == 1
    ), f"There should only be one event with UID {uid} but there are {len(events)}."
    event = events[0]
    inner_text = re.sub(r"\s+", " ", event.get_attribute("innerText"))
    assert inner_text == text, f"Expected {text!r} but got {inner_text!r}"


@when('we click on the event "{text}"')
def step_impl(context, text):
    events = context.browser.find_elements(
        By.XPATH, "//div[contains(@class, ' event ')]"
    )
    chosen_events = [
        event for event in events if text in event.get_attribute("innerText")
    ]
    assert len(chosen_events) == 1, (
        f"There should only be one event with the text {text} "
        f"but there are {len(chosen_events)}: {chosen_events}"
    )
    event = chosen_events[0]
    event.click()


@when('we click on a link containing "{text}"')
def step_impl(context, text):
    event = context.browser.find_element(By.XPATH, f"//a[contains(text(), {text!r})]")
    event.click()


@when('we click on the link "{text}"')
@when('we click the link "{text}"')
def step_impl(context, text):
    links = context.browser.find_elements(By.XPATH, f"//a[text() = {text!r}]")
    assert (
        len(links) == 1
    ), f"I should click on the link {text!r} but found {len(links)}."
    links[0].click()


def get_body_text(context):
    body = context.browser.find_elements(By.XPATH, "//body")[0]
    return body.get_attribute("innerText")


@then('we cannot find an XPATH "{xpath}"')
def step_impl(context, xpath):
    elements = context.browser.find_elements(By.XPATH, xpath)
    for element in elements:
        print(element)
    assert not elements


@then('we cannot see the text "{text}"')
def step_impl(context, text):
    body = get_body_text(context)
    index = body.find(text)
    start = 0 if index < 10 else index - 10
    end = -1 if index > len(body) else index + 10
    assert (
        index == -1
    ), f"{text!r} is visible but should not be visible: {body[start:end]!r}"


@then('we can see the text "{text}"')
def step_impl(context, text):
    assert text in get_body_text(
        context
    ), f"{text!r} is invisible but should be visible"


@then("we can see a {cls}")
def step_impl(context, cls):
    assert context.browser.find_elements(
        By.CLASS_NAME, cls
    ), f"Expected to find elements of class {cls}"


@then('we can see an event with UID "{uid}" with css class "{css_class}"')
def step_impl(context, uid, css_class):
    """Make sure an event has a certain css class."""
    elements_by_uid = context.browser.find_elements(By.CLASS_NAME, f"UID-{uid}")
    elements_by_class = context.browser.find_elements(By.CLASS_NAME, css_class)
    found_elements = set(elements_by_uid) & set(elements_by_class)
    assert found_elements, (
        f"Expected at least one event with UID {uid} "
        f"to have the css class {css_class} but none did."
    )


@then("we cannot see a {cls}")
def step_impl(context, cls):
    assert not context.browser.find_elements(
        By.CLASS_NAME, cls
    ), f"Expected to not find elements of class {cls}"


@when("we open the about page")
def step_impl(context):
    button = context.browser.find_element(By.ID, "infoIcon")
    button.click()
    WebDriverWait(context.browser, 10).until(
        EC.presence_of_element_located((By.ID, "translate-license"))
    )


@when('we select "{dropdown_text}"')
def step_impl(context, dropdown_text):
    options = context.browser.find_elements(
        By.XPATH, f"//option[contains(text(), {dropdown_text!r})]"
    )
    assert options, f"Could not find option with text {dropdown_text!r}"
    if len(options) != 1:
        print(
            f"Expected one option but found many: "
            f"{', '.join(repr(o.get_attribute('innerText')) for o in options)}"
        )
    option = options[0]
    option.click()  # see https://stackoverflow.com/a/7972225/1320237
    wait_for_calendar_to_load(context)


@when("we click on >")
def step_impl(context):
    """Move the calendar one section further."""
    click_button(context, By.CLASS_NAME, "dhx_cal_next_button")


@when("we click on <")
def step_impl(context):
    """Move the calendar one section further."""
    click_button(context, By.CLASS_NAME, "dhx_cal_prev_button")


def click_button(context, selector_type, selector):
    """Find a button with a selector and click it."""
    buttons = context.browser.find_elements(selector_type, selector)
    assert len(buttons) == 1, f"Expected one button but got {buttons}."
    buttons[0].click()


# Browser steps for configuring the calendar.


@given("we are on the configuration page")
def step_impl(context):
    """Visit the configuration page and wait for it to load."""
    context.browser.delete_all_cookies()
    url = context.index_page + "?" + specification_to_query(context.specification)
    get_url(context, url)


@when('we write "{text}" into "{field_id}"')
def step_impl(context, text, field_id):
    """Write text into text input."""
    input_element = context.browser.find_element(By.ID, field_id)
    input_element.clear()  # see https://stackoverflow.com/a/7809907/1320237
    input_element.send_keys(text)
    print(f"Expecting {field_id}.value == {input_element.get_attribute('value')}")


@then('"{text}" is written in "{field_id}"')
def step_impl(context, text, field_id):
    """Check that a field has a value."""
    input_element = context.browser.find_element(By.ID, field_id)
    actual_text = input_element.get_attribute("value")
    assert (
        actual_text == text
    ), f"Expected {text!r} in {field_id} but got {actual_text!r}."


@when('we write the date {day}/{month}/{year} into "{field_id}"')
def step_impl(context, year, month, day, field_id):
    """Write text into text input."""
    input_element = context.browser.find_element(By.ID, field_id)
    input_element.clear()  # see https://stackoverflow.com/a/7809907/1320237
    # construct the text, see https://stackoverflow.com/a/35855868/1320237
    text = context.browser.execute_script(
        f"return (new Date({year}, {month} - 1, {day})).toLocaleDateString()"
    )
    print("text =", repr(text))
    # For filling inputs and date inputs
    # see https://stackoverflow.com/a/35127217/1320237
    # see https://stackoverflow.com/a/39532746/1320237
    ActionChains(context.browser).move_to_element(input_element).click().send_keys(
        text
    ).perform()
    value = input_element.get_attribute("value")
    if not value:
        # this works in Chrome, see https://stackoverflow.com/a/21314269/1320237
        context.browser.execute_script(
            f'document.getElementById({field_id!r}).value = "{year}-{month}-{day}"'
        )
        input_element.send_keys(Keys.CONTROL)
        value = input_element.get_attribute("value")
    print(f"{field_id}.value == {value}")


@when('we choose "{choice}" in "{select_id}"')
def step_impl(context, choice, select_id):
    """Write text into text input."""
    element = context.browser.find_element(By.ID, select_id)
    # see https://stackoverflow.com/a/28613320/1320237
    select = Select(element)
    select.select_by_visible_text(choice)
    print(
        f"{select_id} selected {element.get_attribute('value')!r} "
        f"though text {choice!r}"
    )


def get_specification(context) -> dict:
    """Return the specification from the configuration page."""
    spec_element = context.browser.find_element(By.ID, "json-specification")
    json_string = spec_element.get_attribute("innerText")
    try:
        return json.loads(json_string)
    except:
        print(repr(json_string))
        raise


def assert_specification_has_value(context, attribute, expected_value="no value"):
    """Make sure the specification has a certain value."""
    specification = get_specification(context)
    actual_value = specification.get(attribute, "no value")
    assert (
        actual_value == expected_value
    ), f"specification.{attribute}: expected {expected_value} but got {actual_value}."


@then('"{attribute}" is specified as {expected_value}')
def step_impl(context, attribute, expected_value):
    """Check the JSON value of an attribute."""
    assert_specification_has_value(context, attribute, json.loads(expected_value))


@then('"{attribute}" is not specified')
def step_impl(context, attribute):
    """Check the JSON value of an attribute."""
    assert_specification_has_value(context, attribute)


@when('we click on the {tag:S} "{text}"')
def step_impl(context, tag, text):
    # select if inner text element equals the text
    # see https://stackoverflow.com/a/3655588/1320237
    elements = context.browser.find_elements(By.XPATH, f"//{tag}[text()[. = {text!r}]]")
    if not elements:
        elements = context.browser.find_elements(
            By.XPATH, f"//{tag}[text()[contains(., {text!r})]]"
        )
    assert len(elements) == 1, (
        f"There should only be one {tag} with the text "
        f"{text!r} but there are {len(elements)}."
    )
    element = elements[0]
    element.click()


@then('the checkbox with id "{eid:S}" is checked')
def step_impl(context, eid):
    """Check the checkbox status."""
    element = context.browser.find_element(By.ID, eid)
    assert element.get_attribute("checked")


@then('the checkbox with id "{eid}" is not checked')
def step_impl(context, eid):
    """Check the checkbox status."""
    element = context.browser.find_element(By.ID, eid)
    assert not element.get_attribute("checked")


@when('we click on the {tag_name:S} with id "{eid}"')
def step_impl(context, tag_name, eid):
    """Click on elements with an id."""
    element = context.browser.find_element(By.ID, eid)
    element.click()


## Link verification


@then('the link "{link_text}" targets "{link_target}"')
def step_impl(context, link_text, link_target):
    """Check the target of a link."""
    assert_tag_with_text_attribute_equals(
        context, "a", link_text, "target", link_target
    )


@then('the link "{link_text}" opens "{link_href}"')
def step_impl(context, link_text, link_href):
    """Check the href of a link."""
    assert_tag_with_text_attribute_equals(context, "a", link_text, "href", link_href)


def assert_tag_with_text_attribute_equals(
    context, tag, text, attribute, expected_value
):
    """Make sure that an attribute of a tag has a certain value."""
    # select if inner text element equals the text
    # see https://stackoverflow.com/a/3655588/1320237
    elements = context.browser.find_elements(
        By.XPATH, f"//{tag}[text()[contains(., {text!r})]]"
    )
    assert (
        len(elements) >= 1
    ), f"Expected at least one <{tag}> with text {text!r} but got {len(elements)}."
    actual_values = [element.get_attribute(attribute) for element in elements]
    assert expected_value in actual_values, (
        f"Expected a <{tag}> with the text {text!r} to have an attribute "
        f"{attribute} with the value \n{expected_value!r} but the values are "
        f"\n{{}}.".format("\n".join(map(repr, actual_values)))
    )


## Other



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only
"""Script to convert the DHTMLX calendar translations for yaml files."""

import os
import re

regex = re.compile(r"\{[^{]*(\{.*\})\}", re.DOTALL | re.MULTILINE)

date = "date"
month_full = "month_full:"

en = {}

M = "jan feb mar apr may jun jul aug sep oct nov dec".split()
D = "sun mon tue wed thu fri sat".split()


def used(s):
    return (
        s.startswith("date_")
        or (
            s.startswith("labels_")
            and not any(
                x in s
                for x in "repeat drag message button icon new confirm section "
                "select recurring labels_year_tab labels_grid_tab"
                "labels_full_day".split()
            )
        )
        or s
        in """
    labels_dhx_cal_today_button
    """
    )


def add(l, v):  # noqa: E741
    global yaml  # noqa: PLW0603
    k = "_".join(l)
    # sanitize
    #    v = v.strip()
    assert '"' not in v
    v = '"' + v.replace("=", ":") + '"'
    if v == "":
        v = '""'
    # check translated
    if lang == "en":
        en[k] = v
    elif en.get(k) == v:
        print(f"skip {k} is english: {v}")
        return
    if not used(k):
        #        print(f"Not used {k}")
        yaml += "#: not used in this project\n"
    yaml += f"{k}: {v}\n"


for file in ["locale_en.js"] + os.listdir("."):
    if not file.endswith(".js"):
        continue
    lang = file[file.index("_") + 1 : file.index(".")]
    print(f" ----- {lang} ----- ")
    with open(file) as f:
        r = regex.findall(f.read())
    assert len(r) == 1, r
    r = r[0].replace("{", "dict(").replace("}", ")").replace(":", "=")
    r = eval(r)
    print(list(r))
    print(r)
    yaml = ""
    for x in ["month_full", "month_short"]:
        for k, v in zip(M, r["date"][x]):
            add(["date", x, k], v)
    for x in ["day_full", "day_short"]:
        for k, v in zip(D, r["date"][x]):
            if x == "day_full" and lang == "el" and k == "sat" and v == "Κυριακή":
                # https://github.com/niccokunzmann/open-web-calendar/commit/670b6a7587a3b6bc730301bf288b31afb760d419
                v = "Σάββατο"
            add(["date", x, k], v)
    for k, v in r["labels"].items():
        if not isinstance(v, str):
            assert k in ["month_for_recurring", "day_for_recurring"], v  # checked below
            continue
        add(["labels", k], v)
    if "month_for_recurring" in r["labels"]:
        for k, v in zip(M, r["labels"]["month_for_recurring"]):
            add(["labels", "month_for_recurring", k], v)
    if "day_for_recurring" in r["labels"]:
        for k, v in zip(D, r["labels"]["day_for_recurring"]):
            add(["labels", "day_for_recurring", k], v)
    print(yaml)
    directory = "../../../../translations/" + lang
    os.makedirs(directory, exist_ok=True)
    with open(directory + "/calendar.yml", "w") as f:
        f.write(yaml)



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

import datetime

from flask import Response
from icalendar import Calendar, Event, Timezone
from icalendar.prop import vDDDTypes

from .conversion_base import ConversionStrategy


class ConvertToICS(ConversionStrategy):
    """Convert events to dhtmlx. This conforms to a stratey pattern."""

    def created(self):
        self.title = self.specification["title"]
        self.timezones = set()  # ids

    def is_event(self, component):
        """Whether a component is an event."""
        return isinstance(component, Event)

    def is_timezone(self, component):
        """Whether a component is an event."""
        return isinstance(component, Timezone)

    def collect_components_from(self, calendar_index, calendars):
        for calendar in calendars:
            for component in calendar.walk():
                if self.is_event(component):
                    with self.lock:
                        self.components.append(component)
                if self.is_timezone(component):
                    tzid = component.get("TZID")
                    if tzid and tzid not in self.timezones:
                        with self.lock:
                            self.components.append(component)
                            self.timezones.add(tzid)

    def convert_error(self, error, url, tb_s):
        """Create an error which can be used by the dhtmlx scheduler."""
        event = Event()
        event["DTSTART"] = event["DTEND"] = vDDDTypes(datetime.datetime.now())
        event["SUMMARY"] = type(error).__name__
        event["DESCRIPTION"] = str(error) + "\n\n" + tb_s
        event["UID"] = "error" + str(id(error))
        if url:
            event["URL"] = url
        return event

    def create_calendar(self):
        calendar = Calendar()
        calendar["VERSION"] = "2.0"
        calendar["PRODID"] = "open-web-calendar"
        calendar["CALSCALE"] = "GREGORIAN"
        calendar["METHOD"] = "PUBLISH"
        calendar["X-WR-CALNAME"] = self.title
        calendar["X-PROD-SOURCE"] = self.specification["source_code"]
        return calendar

    def merge(self):
        calendar = self.create_calendar()
        for event in self.components:
            calendar.add_component(event)
        return Response(calendar.to_ical(), mimetype="text/calendar")



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

import io
import sys
import traceback
from concurrent.futures import ThreadPoolExecutor
from threading import RLock
from urllib.parse import urljoin

import requests
from icalendar import Calendar
from lxml import etree


def get_text_from_url(url):
    """Return the text from a url."""
    return requests.get(url, timeout=10).text


class ConversionStrategy:
    """Base class for conversions."""

    # TODO: add as parameters
    MAXIMUM_THREADS = 100

    def __init__(self, specification, get_text_from_url=get_text_from_url):
        self.specification = specification
        self.lock = RLock()
        self.components = []
        self.get_text_from_url = get_text_from_url
        self.created()

    def created(self):
        """Template method for subclasses."""

    def error(self, ty, err, tb, url):
        tb_s = io.StringIO()
        traceback.print_exception(ty, err, tb, file=tb_s)
        return self.convert_error(err, url, tb_s.getvalue())

    def retrieve_calendars(self):
        """Retrieve the calendars from different sources."""
        urls = self.specification["url"]
        if isinstance(urls, str):
            urls = [urls]
        assert len(urls) <= self.MAXIMUM_THREADS, (
            f"You can only merge {self.MAXIMUM_THREADS} urls."
            " If you like more, open an issue."
        )
        with ThreadPoolExecutor(max_workers=self.MAXIMUM_THREADS) as e:
            for _e in e.map(self.retrieve_calendar, enumerate(urls)):
                pass  # no error should pass silently; import this

    def get_calendars_from_url(self, url: str):
        """Return a lis of calendars from a URL."""
        if url.startswith("webcal://"):
            url = url.replace("webcal://", "http://", 1)
        calendar_text = self.get_text_from_url(url)
        try:
            return Calendar.from_ical(calendar_text, multiple=True)
        except ValueError:
            # ValueError: Content line could not be parsed into parts:
            # find the alt link
            # see https://stackoverflow.com/a/11466033/1320237
            htmlparser = etree.HTMLParser()
            tree = etree.XML(calendar_text, htmlparser)
            links = tree.xpath('//link[@rel = "alternate" and @type = "text/calendar"]')
            result = []
            for link in links:
                href = link.get("href")
                new_url = urljoin(url, href)
                calendar_text = self.get_text_from_url(new_url)
                result += Calendar.from_ical(calendar_text, multiple=True)
            if result == []:
                # We did not find any link to any calendar from this source.
                raise
            return result

    def retrieve_calendar(self, index_url):
        """Retrieve a calendar from a url"""
        try:
            index, url = index_url
            calendars = self.get_calendars_from_url(url)
            self.collect_components_from(index, calendars)
        except:
            ty, err, tb = sys.exc_info()
            with self.lock:
                self.components.append(self.error(ty, err, tb, url))

    def collect_components_from(self, index, calendars):
        """Collect all the compenents from the calendar."""
        raise NotImplementedError("to be implemented in subclasses")

    def merge(self):
        """Return the flask Response for the merged calendars."""
        raise NotImplementedError("to be implemented in subclasses")


__all__ = ["ConversionStrategy", "get_text_from_url"]



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only
"""
This test the algorithm written
in the api.md file.
"""

import pytest
from werkzeug.datastructures import MultiDict

from open_web_calendar.app import (
    cache_url,
    get_default_specification,
    get_specification,
)


def test_specification_equals_default_specification_by_default():
    assert get_specification(query=MultiDict({})) == get_default_specification()


def test_specification_overrides_attributes_from_default_specification():
    assert get_default_specification()["css"] == ""
    assert get_specification(query=MultiDict({"css": "test"}))["css"] == "test"


with_url = pytest.mark.parametrize(
    "url",
    [
        "https://tets.io/aksdkkj.json",
        "https://alskdj.asd.de/asld/asdasd.yml",
    ],
)


@with_url
def test_specification_prefers_url_over_default(url):
    cache_url(url, '{"css": "123"}')
    assert (
        get_specification(query=MultiDict({"specification_url": url}))["css"] == "123"
    )


@with_url
def test_url_parameters_are_more_important_than_specification_url(url):
    cache_url(url, '{"test": "123"}')
    assert (
        get_specification(query=MultiDict({"specification_url": url, "test": "test"}))[
            "test"
        ]
        == "test"
    )


@with_url
def test_specification_can_be_loaded_from_yml_files(url):
    cache_url(url, 'test: "123"')
    assert (
        get_specification(query=MultiDict({"specification_url": url}))["test"] == "123"
    )


@pytest.mark.parametrize(
    ("string", "expected_bool"),
    [
        ("true", True),
        ("True", True),
        ("false", False),
        ("False", False),
    ],
)
def test_false_and_true(string, expected_bool):
    """Check that True and False work."""
    value = get_specification(query=MultiDict({"param": string}))["param"]
    assert bool(value) == bool(expected_bool)
    assert (not value) == (not expected_bool)


def test_boolean_values_of_parameters():
    """Some parameters have boolean values."""
    spec = get_specification(
        query=MultiDict({"clean_html_style": "false", "clean_html_links": "True"})
    )
    assert spec["clean_html_embedded"] is True, "default"
    assert spec["clean_html_style"] is False
    assert spec["clean_html_links"] is True



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only
from __future__ import annotations

import datetime
from html import escape
from urllib.parse import unquote

import pytz
import recurring_ical_events
from dateutil.parser import parse as parse_date
from flask import jsonify

from .clean_html import clean_html
from .conversion_base import ConversionStrategy


def is_date(date):
    """Whether the date is a datetime.date and not a datetime.datetime"""
    return isinstance(date, datetime.date) and not isinstance(date, datetime.datetime)


class ConvertToDhtmlx(ConversionStrategy):
    """Convert events to dhtmlx. This conforms to a stratey pattern.

    - timeshift_minutes is the timeshift specified by the calendar
        for dates.
    """

    def created(self):
        """Set attribtues when created."""
        try:
            self.timezone = pytz.timezone(self.specification["timezone"])
        except pytz.UnknownTimeZoneError:
            self.timezone = pytz.FixedOffset(-int(self.specification["timeshift"]))
        self.today = today = (
            parse_date(self.specification["date"])
            if self.specification.get("date")
            else datetime.datetime.now(self.timezone)
        )
        self.to_date = (
            parse_date(self.specification["to"])
            if self.specification.get("to")
            else today.replace(year=today.year + 1)
        )
        self.from_date = (
            parse_date(self.specification["from"])
            if self.specification.get("from")
            else today.replace(year=today.year - 1)
        )

    def date_to_string(self, date):
        """Convert a date to a string."""
        # use ISO format
        # see https://docs.dhtmlx.com/scheduler/howtostart_nodejs.html#step4implementingcrud
        # see https://docs.python.org/3/library/datetime.html#datetime.datetime.isoformat
        # see https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
        if is_date(date):
            date = datetime.datetime(
                date.year, date.month, date.day, tzinfo=self.timezone
            )
        elif date.tzinfo is None:
            date = self.timezone.localize(date)
        # convert to other timezone, see https://stackoverflow.com/a/54376154
        viewed_date = date.astimezone(self.timezone)
        return viewed_date.strftime("%Y-%m-%d %H:%M")

    def convert_ical_event(self, calendar_index, calendar_event):
        start = calendar_event["DTSTART"].dt
        end = calendar_event.get("DTEND", calendar_event["DTSTART"]).dt
        if is_date(start) and is_date(end) and end == start:
            end = datetime.timedelta(days=1) + start
        geo = calendar_event.get("GEO", None)
        if geo:
            geo = {"lon": geo.longitude, "lat": geo.latitude}
        name = calendar_event.get("SUMMARY", "")
        sequence = str(calendar_event.get("SEQUENCE", 0))
        uid = calendar_event.get(
            "UID", ""
        )  # issue 69: UID is helpful for debugging but not required
        start_date = self.date_to_string(start)
        return {
            "start_date": start_date,
            "end_date": self.date_to_string(end),
            "start_date_iso": start.isoformat(),
            "end_date_iso": end.isoformat(),
            "start_date_iso_0": start.isoformat(),
            "end_date_iso_0": end.isoformat(),
            "text": name,
            "description": self.get_event_description(calendar_event),
            "location": calendar_event.get("LOCATION", None),
            "geo": geo,
            "uid": uid,
            "ical": calendar_event.to_ical().decode("UTF-8"),
            "sequence": sequence,
            "recurrence": None,
            "url": calendar_event.get("URL"),
            "id": uid + "-" + start_date.replace(" ", "-").replace(":", "-"),
            "type": "event",
            "color": calendar_event.get(
                "COLOR", calendar_event.get("X-APPLE-CALENDAR-COLOR", "")
            ),
            "categories": self.get_event_categories(calendar_event),
            "css-classes": ["event"]
            + self.get_event_classes(calendar_event)
            + [f"CALENDAR-INDEX-{calendar_index}"],
        }

    def convert_error(self, error, url, tb_s):
        """Create an error which can be used by the dhtmlx scheduler."""
        # always add the error within the requested time range
        now = self.from_date
        now_iso = now.isoformat()
        now_s = self.date_to_string(now)
        return {
            "start_date": now_s,
            "end_date": now_s,
            "start_date_iso": now_iso,
            "end_date_iso": now_iso,
            "start_date_iso_0": now_iso,
            "end_date_iso_0": now_iso,
            "text": type(error).__name__,
            "description": self.clean_html(escape(str(error))),
            "traceback": self.clean_html(escape(tb_s)),
            "location": None,
            "geo": None,
            "uid": "error",
            "ical": "",
            "sequence": 0,
            "recurrence": None,
            "url": url,
            "id": id(error),
            "type": "error",
            "css-classes": ["error"],
        }

    def get_event_description(self, event):
        """Return a formatted description of the event.

        HTML is cleaned.
        """
        # CMS4Schools.com
        description = event.get("X-ALT-DESC")
        if (
            description is None
            or not hasattr(description, "params")
            or description.params.get("FMTTYPE") != "text/html"
        ):
            description = event.get("DESCRIPTION", "")
            if hasattr(description, "params"):
                altrep = description.params.get("ALTREP")
                if altrep is not None and "," in altrep:
                    data, content = altrep.split(",", 1)
                    if data == "data:text/html":
                        # Thunderbird html
                        description = unquote(content)
        return self.clean_html(description)

    def clean_html(self, html):
        """Return the cleaned HTML."""
        return clean_html(html, self.specification)

    def merge(self):
        return jsonify(self.components)

    def collect_components_from(self, calendar_index, calendars):
        # see https://stackoverflow.com/a/16115575/1320237
        for calendar in calendars:
            events = recurring_ical_events.of(calendar).between(
                self.from_date, self.to_date
            )
            with self.lock:
                for event in events:
                    json_event = self.convert_ical_event(calendar_index, event)
                    self.components.append(json_event)

    def get_event_classes(self, event) -> list[str]:
        """Return the CSS classes that should be used for the event styles."""
        classes = []
        for attr in ["UID", "TRANSP", "STATUS", "CLASS", "PRIORITY"]:
            value = event.get(attr)
            if value is not None:
                classes.append(f"{attr}-{value}")
        if event.get("CLASS") not in [None, "PUBLIC", "CONFIDENTIAL", "PRIVATE"]:
            classes.append("CLASS-PRIVATE")  # unrecognized is private
        for category in self.get_event_categories(event):
            classes.append(f"CATEGORY-{category}")
        return classes

    def get_event_categories(self, event) -> list[str]:
        """Return the categories of the event."""
        categories = event.get("CATEGORIES", None)
        return categories.cats if categories is not None else []


__all__ = ["ConvertToDhtmlx"]



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Clean HTML content of events.

This is important to mitigate attacks from ICS sources.
See https://stackoverflow.com/questions/3073881/clean-up-html-in-python
"""

import re
import warnings

from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning
from lxml.html.clean import Cleaner

CLEAN_HTML_SPECIFICATION_PREFIX = "clean_html_"
DEFAULT_SPEC = {
    "page_structure": True,
    "remove_tags": ("body", "div"),
}
HTML_TAG_MATCH = re.compile("<[^>]*>")


def clean_html(bad_html: str, spec: dict) -> str:
    """Clean up the HTML.

    For the content of the spec parameter, see
    - the default_specification.yml file, clean_html_* attributes
    - https://lxml.de/api/lxml.html.clean.Cleaner-class.html
    """
    with warnings.catch_warnings():
        # ignore that input might be short
        # see https://stackoverflow.com/a/17654868/1320237
        warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)
        tree = BeautifulSoup(bad_html, "html.parser")
    bad_html = tree.prettify()
    kw = DEFAULT_SPEC.copy()
    kw.update(
        {
            name[len(CLEAN_HTML_SPECIFICATION_PREFIX) :]: value
            for name, value in spec.items()
            if name.startswith(CLEAN_HTML_SPECIFICATION_PREFIX)
        }
    )
    cleaner = Cleaner(**kw)
    result = cleaner.clean_html(f"<div>{bad_html}</div>").strip()
    if result.startswith("<div>"):
        result = result[5:-6]
    return result


def remove_html(html: str) -> str:
    """Remove all HTML from the html string and only return the text."""
    return HTML_TAG_MATCH.sub("", html)


__all__ = ["clean_html", "remove_html"]



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

import io
import json
import os
import tempfile
import traceback
from pathlib import Path

import pytz
import requests
import yaml
from flask import (
    Flask,
    Response,
    jsonify,
    make_response,
    render_template,
    request,
    send_from_directory,
)
from flask_allowed_hosts import AllowedHosts
from flask_caching import Cache

from . import translate, version
from .convert_to_dhtmlx import ConvertToDhtmlx
from .convert_to_ics import ConvertToICS

# configuration
DEBUG = os.environ.get("APP_DEBUG", "true").lower() == "true"
PORT = int(os.environ.get("PORT", "5000"))
CACHE_REQUESTED_URLS_FOR_SECONDS = int(
    os.environ.get("CACHE_REQUESTED_URLS_FOR_SECONDS", 600)
)
ALLOWED_HOSTS = os.environ.get("ALLOWED_HOSTS", "").split(",")
if ALLOWED_HOSTS == [""]:  # noqa: SIM300, RUF100
    ALLOWED_HOSTS = []
REQUESTS_TIMEOUT = int(os.environ.get("SOURCE_TIMEOUT", "60"))

# constants
HERE = Path(__file__).parent
DEFAULT_SPECIFICATION_PATH = HERE / "default_specification.yml"
TEMPLATE_FOLDER_NAME = "templates"
TEMPLATE_FOLDER = HERE / TEMPLATE_FOLDER_NAME
CALENDARS_TEMPLATE_FOLDER_NAME = "calendars"
CALENDAR_TEMPLATE_FOLDER = TEMPLATE_FOLDER / CALENDARS_TEMPLATE_FOLDER_NAME
STATIC_FOLDER_NAME = "static"
STATIC_FOLDER_PATH = HERE / STATIC_FOLDER_NAME
DEFAULT_REQUEST_HEADERS = {
    "user-agent": "open-web-calendar",
}

# specification
PARAM_SPECIFICATION_URL = "specification_url"

# globals
app = Flask(__name__, template_folder="templates")
# Check Configuring Flask-Cache section for more details
CACHE_CONFIG = {
    "CACHE_TYPE": "FileSystemCache",
    "CACHE_DIR": tempfile.mkdtemp(prefix="owc-cache-"),
}
cache = Cache(app, config=CACHE_CONFIG)

# caching

__URL_CACHE = {}


# limiting access
def host_not_allowed():
    return render_template(
        "403.html",
        hostname=request.host.split(":")[0],
        allowed_hosts=", ".join(ALLOWED_HOSTS),
    ), 403


allowed_hosts = AllowedHosts(
    app, allowed_hosts=ALLOWED_HOSTS, on_denied=host_not_allowed
)

# This is an in-app override of the default_specification.yml
DEFAULT_SPECIFICATION = {}


def cache_url(url, text):
    """Cache the value of a url."""
    __URL_CACHE[url] = text
    try:
        get_text_from_url(url)
    finally:
        del __URL_CACHE[url]


@app.after_request
def add_header(r):
    """
    Add headers to both force latest IE rendering engine or Chrome Frame,
    and also to cache the rendered page for 10 minutes.
    """
    r.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
    r.headers["Pragma"] = "no-cache"
    r.headers["Expires"] = "0"
    return r


# configuration


def get_configuration():
    """Return the configuration for the browser."""
    return {
        "default_specification": get_default_specification(),
        "version": version.version,
        "version-list": version.version_tuple,
        "timezones": pytz.all_timezones,  # see https://stackoverflow.com/a/13867319
        "dhtmlx": {"languages": translate.dhtmlx_languages()},
        "index": {"languages": translate.languages_for_the_index_file()},
    }


def set_js_headers(response):
    response = make_response(response)
    response.headers["Access-Control-Allow-Origin"] = "*"
    # see https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS/Errors/CORSMissingAllowHeaderFromPreflight
    response.headers["Access-Control-Allow-Headers"] = request.headers.get(
        "Access-Control-Request-Headers"
    )
    if "Content-Type" not in response.headers:
        response.headers["Content-Type"] = "text/calendar"
    return response


def make_js_file_response(content: str) -> Response:
    """Modify the response to set the content type for .js files."""
    response = make_response(content)
    set_js_headers(response)
    response.headers["Content-Type"] = "text/javascript"
    return response


@cache.memoize(
    CACHE_REQUESTED_URLS_FOR_SECONDS, forced_update=lambda: bool(__URL_CACHE)
)
def get_text_from_url(url):
    """Return the text from a url.

    The result is cached CACHE_REQUESTED_URLS_FOR_SECONDS.
    """
    if __URL_CACHE:
        return __URL_CACHE[url]
    return requests.get(
        url, headers=DEFAULT_REQUEST_HEADERS, timeout=REQUESTS_TIMEOUT
    ).content


def get_default_specification():
    """Return the default specification."""
    with DEFAULT_SPECIFICATION_PATH.open(encoding="UTF-8") as file:
        spec = yaml.safe_load(file)
        spec.update(DEFAULT_SPECIFICATION)
        return spec


def get_specification(query=None):
    """Build the calendar specification."""
    if query is None:
        query = request.args
    specification = get_default_specification()
    # get a request parameter, see https://stackoverflow.com/a/11774434
    url = query.get(PARAM_SPECIFICATION_URL, None)
    if url:
        url_specification_response = get_text_from_url(url)
        try:
            url_specification_values = json.loads(url_specification_response)
        except json.JSONDecodeError:
            url_specification_values = yaml.safe_load(url_specification_response)
        specification.update(url_specification_values)
    for parameter in query:
        # get a list of arguments
        # see https://web.archive.org/web/20230325034825/https://werkzeug.palletsprojects.com/en/0.14.x/datastructures/
        value = query.getlist(parameter)
        # convert values
        for i in range(len(value)):
            if value[i] in ("false", "False"):
                value[i] = False
            elif value[i] in ("true", "True"):
                value[i] = True
        if len(value) == 1 and not isinstance(specification.get(parameter), list):
            value = value[0]
        specification[parameter] = value

    return specification


def get_query_string():
    return "?" + request.query_string.decode()


def render_app_template(template, specification):
    translation_file = Path(template).stem
    language = specification["language"]
    if specification["prefer_browser_language"]:
        # see https://stackoverflow.com/a/30441752/1320237
        # see https://tedboy.github.io/flask/generated/generated/werkzeug.LanguageAccept.html
        language = request.accept_languages.best_match(
            translate.LANGUAGE_CODES, language
        )
    return render_template(
        template,
        specification=specification,
        configuration=get_configuration(),
        json=json,
        get_query_string=get_query_string,
        html=lambda tid, **template_replacements: translate.html(
            language, translation_file, tid, **template_replacements
        ),
        language=language,
    )


@app.route("/calendar.<ext>", methods=["GET", "OPTIONS"])
@allowed_hosts.limit()
# use query string in cache, see https://stackoverflow.com/a/47181782/1320237
# @cache.cached(timeout=CACHE_TIMEOUT, query_string=True)
def get_calendar(ext):
    """Return a calendar."""
    specification = get_specification()
    if ext == "spec":
        return jsonify(specification)
    if ext == "events.json":
        strategy = ConvertToDhtmlx(specification, get_text_from_url)
        strategy.retrieve_calendars()
        return set_js_headers(strategy.merge())
    if ext == "ics":
        strategy = ConvertToICS(specification, get_text_from_url)
        strategy.retrieve_calendars()
        return set_js_headers(strategy.merge())
    if ext == "html":
        template_name = specification["template"]
        all_template_names = os.listdir(CALENDAR_TEMPLATE_FOLDER)
        assert (
            template_name in all_template_names
        ), 'Template names must be file names like "{}", not "{}".'.format(
            '", "'.join(all_template_names), template_name
        )
        template = CALENDARS_TEMPLATE_FOLDER_NAME + "/" + template_name
        return render_app_template(template, specification)
    raise ValueError(
        f"Cannot use extension {ext}. Please see the documentation or report an error."
    )


for folder_path in STATIC_FOLDER_PATH.iterdir():
    if not folder_path.is_dir():
        continue

    @app.route(
        "/" + folder_path.name + "/<path:path>", endpoint="static/" + folder_path.name
    )
    def send_static(path, folder_name=folder_path.name):
        return send_from_directory("static/" + folder_name, path)


@app.route("/")
@app.route("/index.html")
@allowed_hosts.limit()
def serve_index():
    specification = get_specification()
    return render_app_template("index.html", specification)


@app.route("/about.html")
@allowed_hosts.limit()
def serve_about():
    specification = get_specification()
    return render_app_template("about.html", specification)


@app.route("/configuration.js")
@allowed_hosts.limit()
def serve_configuration():
    return make_js_file_response(
        f"/* generated */\nconst configuration = {json.dumps(get_configuration())};"
    )


@app.route("/locale_<lang>.js")
@allowed_hosts.limit()
def serve_locale(lang):
    """Serve the locale translations for the web frontend DHTMLX."""
    return make_js_file_response(
        render_template(
            "locale.js", locale=json.dumps(translate.dhtmlx(lang), indent="  ")
        )
    )


@app.errorhandler(500)
def unhandled_exception(error):
    """Called when an error occurs.

    See https://stackoverflow.com/q/14993318
    """
    file = io.StringIO()
    traceback.print_exception(type(error), error, error.__traceback__, file=file)
    return (
        f"""
    <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
    <html>
        <head>
            <title>500 Internal Server Error</title>
        </head>
        <body>
            <h1>Internal Server Error</h1>
            <p>
                The server encountered an internal error and was unable to
                complete your request.  Either the server is overloaded or
                there is an error in the application.
            </p>
            <pre>\r\n{file.getvalue()}
            </pre>
        </body>
    </html>
    """,
        500,
    )  # return error code from https://stackoverflow.com/a/7824605


# make serializable for multiprocessing
# app.__reduce__ = lambda: __name__ + ".app"


def main():
    """Run the Open Web Calendar"""
    print("""If you want to run the Open Web Calendar in production,
please use this command:

    gunicorn open_web_calendar:app
    """)  # noqa: T201

    app.run(debug=DEBUG, host="0.0.0.0", port=PORT)


__all__ = [
    "main",
    "app",
    "DEFAULT_SPECIFICATION",
    "cache_url",
    "get_text_from_url",
    "get_default_specification",
    "get_specification",
    "DEFAULT_REQUEST_HEADERS",
]

if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Test the headers of responses."""

import pytest

CAL_JSON = "/calendar.events.json"
CAL_ICS = "/calendar.ics"


def test_json_result(client):
    """Check the JS headers."""
    response = client.get(CAL_JSON)
    assert response.access_control_allow_origin == "*"
    assert response.content_type.startswith("application/json")


def test_ics(client):
    """Check the JS headers."""
    response = client.get(CAL_ICS)
    print(dir(response))
    assert response.access_control_allow_origin == "*"
    assert response.content_type.startswith("text/calendar")


@pytest.mark.parametrize("endpoint", [CAL_ICS, CAL_JSON])
@pytest.mark.parametrize("h", ["", "asd asd2"])
def test_allow_headers(endpoint, client, h):
    """Check the allowed headers"""
    response = client.get(CAL_ICS, headers={"Access-Control-Request-Headers": h})
    print(response.text)
    assert response.headers.get("Access-Control-Allow-Headers") == h


@pytest.mark.parametrize("endpoint", [CAL_ICS, CAL_JSON])
def test_return_code(client, endpoint):
    """Check the return code"""
    result = client.get(endpoint)
    print(result, dir(result))
    assert result.status_code == 200



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

import sys
from pathlib import Path
from unittest.mock import Mock

import pytest
import requests

# constants
HERE = Path(__file__).parent
CALENDAR_DIRECTORY = HERE / ".." / "features" / "calendars"

# relative imports
sys.path.append(HERE.absolute() / ".." / "..")
sys.path.append(HERE.absolute())
from open_web_calendar.app import DEFAULT_SPECIFICATION, cache_url  # noqa: E402

DEFAULT_SPECIFICATION["url"] = []


@pytest.fixture(autouse=True)
def _no_requests(monkeypatch):
    """Prevent requests from sending out requests

    See https://docs.pytest.org/en/latest/monkeypatch.html#example-preventing-requests-from-remote-operations
    """

    def test_cannot_call_outside(*args, **kw):
        raise RuntimeError(
            "Tests are not allowed to make requests to the"
            " Internet. You can use cache_url() to mock that."
        )

    monkeypatch.setattr(requests.sessions.Session, "request", test_cannot_call_outside)


@pytest.fixture
def mock():
    return Mock()


@pytest.fixture
def app():
    """Create the app.

    See https://flask.palletsprojects.com/en/2.2.x/testing/
    """
    from open_web_calendar.app import app

    app.config.update(
        {
            "TESTING": True,
        }
    )

    # other setup can go here

    return app

    # clean up / reset resources here


@pytest.fixture
def client(app):
    return app.test_client()


@pytest.fixture
def runner(app):
    return app.test_cli_runner()


calendar_files = {}
for file in CALENDAR_DIRECTORY.iterdir():
    with file.open() as f:
        calendar_files[file] = f.read()


@pytest.fixture
def calendar_urls():
    """Mapping the calendar name without .ics to the cached url.

    The files are located in the CALENDAR_FOLDER.
    """
    mapping = {}
    for file, content in calendar_files.items():
        url = "http://test.examples.local/" + file.name
        cache_url(url, content)
        mapping[file.name] = url
        if file.suffix.lower() == ".ics":
            mapping[file.stem] = url
    return mapping


@pytest.fixture
def calendar_content():
    """Mapping the calendar name without .ics to the calendar content.

    The files are located in the CALENDAR_FOLDER.
    """
    mapping = {}
    for file, content in calendar_files.items():
        if file.suffix.lower() == ".ics":
            mapping[file.stem] = content
    return mapping



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

from .app import main

main()



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Tests for Issue 241.

See https://github.com/niccokunzmann/open-web-calendar/issues/241

We expect text/javascript.
https://stackoverflow.com/a/189877/1320237
"""

import pytest


@pytest.mark.parametrize(
    "js_url",
    [
        "/configuration.js",
        "/js/index.js",
        "/locale_de.js",
        "/locale_nb_NO.js",
    ],
)
def test_mime_type_of_configuration_js(client, js_url):
    """Check the Content-Type header."""
    response = client.get(js_url)
    assert "text/javascript" in response.content_type.split(";")



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only
import pytest

from open_web_calendar import translate


@pytest.mark.parametrize(
    ("language", "file", "tid", "expected_value"),
    [
        ("en", "calendar", "language", "English"),
        ("en", "_test", "id1", "test"),
        ("en", "_test", "id2", "test2"),
        ("de", "_test", "id1", "Test"),
        ("de", "_test", "id2", "test2"),
        ("en", "_test", "calendar.language", "English"),
    ],
)
def test_translate(language, file, tid, expected_value):
    """Translate the id."""
    assert translate.string(language, file, tid) == expected_value


@pytest.mark.parametrize(
    ("language", "tid", "value", "fill"),
    [
        ("en", "id2", '<span id="translate-id2" class="translation">test2</span>', {}),
        (
            "en",
            "test-html",
            '<span id="translate-test-html" class="translation"><b>Hi!</b></span>',
            {},
        ),
        (
            "en",
            "check-escape",
            '<span id="translate-check-escape" class="translation">&lt;nanana&gt;</span>',
            {},
        ),
        (
            "en",
            "format-argument",
            '<span id="translate-format-argument" class="translation">This is a nice day!</span>',
            {"sunny": "nice"},
        ),
        (
            "en",
            "format-argument",
            '<span id="translate-format-argument" class="translation">This is a rainy day!</span>',
            {"sunny": "rainy"},
        ),
    ],
)
def test_convert_html(language, tid, value, fill):
    """Check that html conversion works."""
    assert translate.html(language, "_test", tid, **fill) == value


@pytest.mark.parametrize("language", ["en", "de"])
def test_load_from_common(language):
    assert translate.string(
        language, "common", "project-description"
    ) == translate.string(language, "index", "project-description")


def test_invalid_id():
    with pytest.raises(KeyError):
        translate.string("en", "common", "invalid-id-for-test")


def test_ua_calendar():
    """Check that we get a valid dhtmlx scheduler translation."""
    expected = {
        "date": {
            "month_full": [
                "Січень",
                "Лютий",
                "Березень",
                "Квітень",
                "Травень",
                "Червень",
                "Липень",
                "Серпень",
                "Вересень",
                "Жовтень",
                "Листопад",
                "Грудень",
            ],
            "month_short": [
                "Січ",
                "Лют",
                "Бер",
                "Кві",
                "Тра",
                "Чер",
                "Лип",
                "Сер",
                "Вер",
                "Жов",
                "Лис",
                "Гру",
            ],
            "day_full": [
                "Неділя",
                "Понеділок",
                "Вівторок",
                "Середа",
                "Четвер",
                "П'ятниця",
                "Субота",
            ],
            "day_short": ["Нед", "Пон", "Вів", "Сер", "Чет", "Птн", "Суб"],
        },
        "labels": {
            "dhx_cal_today_button": "Сьогодні",
            "day_tab": "День",
            "week_tab": "Тиждень",
            "month_tab": "Місяць",
            "new_event": "Нова подія",
            "icon_save": "Зберегти",
            "icon_cancel": "Відміна",
            "icon_details": "Деталі",
            "icon_edit": "Редагувати",
            "icon_delete": "Вилучити",
            "confirm_closing": "",
            "confirm_deleting": "Подія вилучиться назавжди. Ви впевнені?",
            "section_description": "Опис",
            "section_time": "Часовий проміжок",
            "full_day": "Весь день",
            "confirm_recurring": "Хочете редагувати весь перелік повторюваних подій?",
            "section_recurring": "Повторювана подія",
            "button_recurring": "Відключено",
            "button_recurring_open": "Включено",
            "button_edit_series": "Редагувати серію",
            "button_edit_occurrence": "Редагувати примірник",
            "agenda_tab": "Перелік",
            "date": "Дата",
            "description": "Опис",
            "year_tab": "Рік",
            "week_agenda_tab": "Перелік",
            "grid_tab": "Таблиця",
            "drag_to_create": "Drag to create",
            "drag_to_move": "Drag to move",
            "message_ok": "OK",
            "message_cancel": "Cancel",
            "next": "Наступний",
            "prev": "Попередній",
            "year": "Рік",
            "month": "Місяць",
            "day": "День",
            "hour": "Година",
            "minute": "Хвилина",
            "repeat_radio_day": "День",
            "repeat_radio_week": "Тиждень",
            "repeat_radio_month": "Місяць",
            "repeat_radio_year": "Рік",
            "repeat_radio_day_type": "Кожний",
            "repeat_text_day_count": "день",
            "repeat_radio_day_type2": "Кожний робочий день",
            "repeat_week": " Повторювати кожен",
            "repeat_text_week_count": "тиждень , по:",
            "repeat_radio_month_type": "Повторювати",
            "repeat_radio_month_start": "",
            "repeat_text_month_day": " числа кожний ",
            "repeat_text_month_count": "місяць",
            "repeat_text_month_count2_before": "кожен ",
            "repeat_text_month_count2_after": "місяць",
            "repeat_year_label": "",
            "select_year_day2": "",
            "repeat_text_year_day": "день",
            "select_year_month": "",
            "repeat_radio_end": "Без дати закінчення",
            "repeat_text_occurences_count": "повторень",
            "repeat_radio_end3": "До ",
            "repeat_radio_end2": "",
            "month_for_recurring": [
                "січня",
                "лютого",
                "березня",
                "квітня",
                "травня",
                "червня",
                "липня",
                "серпня",
                "вересня",
                "жовтня",
                "листопада",
                "грудня",
            ],
            "day_for_recurring": [
                "Неділям",
                "Понеділкам",
                "Вівторкам",
                "Середам",
                "Четвергам",
                "П'ятницям",
                "Суботам",
            ],
        },
    }
    cal = translate.dhtmlx("ua")

    def compare(expected, b, gone=None):
        if gone is None:
            gone = []
        for k in expected:
            assert k in b
            if isinstance(expected[k], (list, str)):
                assert expected[k] == b[k], ".".join(gone + [k])
            elif isinstance(expected[k], dict):
                compare(expected[k], b[k], gone + [k])
            else:
                assert False

    compare(expected, cal)


@pytest.mark.parametrize(
    "entry",
    [
        ("Deutsch", "de"),
        ("English", "en"),
        ("Portuguese", "pt"),
    ],
)
def test_languages_are_listed(entry):
    languages = translate.dhtmlx_languages()
    assert entry in languages


def test_language_alias_is_not_a_translation_language():
    """If this fails, please edit the file to rename language codes properly."""
    for code in translate.LANGUAGE_ALIAS:
        assert code not in translate.TRANSLATIONS


def test_compute_recommendation():
    assert translate.languages_for_the_index_file() != []


@pytest.mark.parametrize("code", ["nb-NO", "nb-no", "nb_NO"])
def test_language_code_with_minus_is_ok(code):
    """We are also allowed to use the - in language codes."""
    assert translate.string(code, "calendar", "labels_minute") == "Minutt"


@pytest.mark.parametrize("code", ["de-DE", "de-AU", "de_CH", "en-US", "en_US"])
def test_language_falls_back_to_base(code):
    """We are also allowed to use the - in language codes."""
    default_code = code.split("-")[0].split("_")[0]
    default = translate.string(default_code, "calendar", "language")
    expected = translate.string(code, "calendar", "language")
    assert (
        default == expected
    ), "The default code should be the same if the language is not given."


@pytest.mark.parametrize(
    ("code", "codes"),
    [
        ("en", ["en"]),
        ("en_US", ["en_US", "en"]),
        ("de-AU", ["de_AU", "de", "en"]),
        ("zh-CN", ["zh_Hans", "zh", "en"]),
    ],
)
def test_language_list(code, codes):
    """Check that we generate the right lookup order."""
    assert translate.get_language_lookup(code) == codes


def test_languages_codes():
    """We want to make sure all languages can be matched."""
    assert "en" in translate.LANGUAGE_CODES
    assert "de" in translate.LANGUAGE_CODES
    assert "cn" in translate.LANGUAGE_CODES, "alial occurs"
    assert "nb-NO" in translate.LANGUAGE_CODES, "code with - occurs"



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only
"""This file is an interface file for comptibility.

Originally, open_web_calendar/app.py was in this place.
To provide the same interface, this file was created.
"""

from open_web_calendar import app, main

__all__ = ["app", "main"]

if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""This tests adding an alternative link to the ICS file.

The link can be relative or absolute.

See https://github.com/niccokunzmann/open-web-calendar/issues/308
"""

import html

import pytest

from open_web_calendar.app import cache_url


@pytest.mark.parametrize("page", ["index.html", "about.html", "calendar.html"])
@pytest.mark.parametrize(
    "query",
    [
        "a=1&b=2",
        "specification_url=https://open-web-calendar.quelltext.eu/assets/templates/free-and-busy.json",
    ],
)
def test_the_query_is_used_in_the_alternate_link(client, page, query):
    """Make sure an alternate link is present."""
    cache_url(
        "https://open-web-calendar.quelltext.eu/assets/templates/free-and-busy.json",
        "{}",
    )
    response = client.get(f"/{page}?{query}")
    # html.escape, see https://stackoverflow.com/a/5072031/1320237
    link = f'<link rel="alternate" type="text/calendar" href="/calendar.ics?{html.escape(query)}" />'
    print(response.text[:1000])
    assert link in response.text



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Check that we can open webcal links.

webcal:// usually means http.
"""

from open_web_calendar.conversion_base import ConversionStrategy


def test_webcal_is_replaced(mock):
    """Check that when we use webcal://, we actually use http"""
    cb = ConversionStrategy({}, mock)
    mock.return_value = ""
    cb.get_calendars_from_url("webcal://url.to/a/calendar.ics")
    mock.assert_called_once_with("http://url.to/a/calendar.ics")



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only
"""Set the canonical URL of pages

See https://www.mkdocs.org/user-guide/configuration/#hooks
See https://github.com/squidfunk/mkdocs-material/discussions/6730#discussioncomment-8339285

This makes the links easily detectable by the link checker.
"""

CANONICAL_EXT = ".canonical"


def on_page_markdown(markdown, *, page, config, files):
    if not page.canonical_url.endswith(CANONICAL_EXT):
        page.canonical_url = page.canonical_url + CANONICAL_EXT



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Use the HTML description where given.

See https://github.com/niccokunzmann/open-web-calendar/issues/300

Calendars:
- description is HTML: issue-287-links-1.ics
- description parameter is HTML: event-with-html-markup.ics
- alt description attribtue is HTML: food.ics

"""

from icalendar import Calendar, Event

from open_web_calendar.app import ConvertToDhtmlx


def event_with_uid(calendar_content: str, uid: str) -> Event:
    """Return the first event with the UID."""
    calendar = Calendar.from_ical(calendar_content)
    events = [event for event in calendar.walk("VEVENT") if event["UID"] == uid]
    assert events
    return events[0]


def test_description_is_html(calendar_content):
    """The description of the event is already in HTML."""
    event = event_with_uid(
        calendar_content["issue-287-links-1"], "1jkma74p98bj79c760660arc07@google.com"
    )
    description = ConvertToDhtmlx({"timezone": "Europe/London"}).get_event_description(
        event
    )
    assert '<a href="https://bethanymarceline.com/"' in description


def test_description_is_parameter(calendar_content):
    """The description of the event is already in HTML."""
    event = event_with_uid(
        calendar_content["event-with-html-markup"],
        "683642b3-9b25-4177-8b46-ec2f65e64020",
    )
    description = ConvertToDhtmlx({"timezone": "Europe/London"}).get_event_description(
        event
    )
    assert "<h1>\n Know This Heading!\n</h1>" in description


def test_alt_attribute(calendar_content):
    """The HTML description is in an alt attribute."""
    event = event_with_uid(calendar_content["food"], "2851")
    description = ConvertToDhtmlx({"timezone": "Europe/London"}).get_event_description(
        event
    )
    assert "Cauliflower\n  <br>" in description



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Translations of the program into other languages."""

from __future__ import annotations

import itertools
import os
from collections import defaultdict
from html import escape
from pathlib import Path

import yaml
from markupsafe import Markup

HERE = Path(__file__).parent
TRANSLATIONS_PATH = HERE / "translations"

# The default language is that of the translation system look-up
# and not to be confused with the language given in the
# default_specificaition.yml.
DEFAULT_LANGUAGE = "en"
DEFAULT_FILE = "common"
CALENDAR_FILE = "calendar"
TRANSLATIONS = {}  # lang -> file -> id -> string
LANGUAGE_ALIAS = {  # name also usable -> name in the translations directory
    "nb": "nb_NO",
    "ua": "uk",
    "jp": "ja",
    "cn": "zh_Hans",
    "no": "nb_NO",
    "zh_CN": "zh_Hans",  # use _!
}  # rename language codes
UNUSED = "-unused"

# Parse the directory structure.
# translations/<lang>/<file>.yml
for language in os.listdir(TRANSLATIONS_PATH):
    TRANSLATIONS[language] = file_translations = defaultdict(dict)
    language_path = TRANSLATIONS_PATH / language
    for file in language_path.iterdir():
        if file.suffix != ".yml":
            continue
        name = file.stem
        if name.endswith(UNUSED):
            name = name[: -len(UNUSED)]
        with file.open() as f:
            file_translations[name].update(yaml.safe_load(f))


def get_language_lookup(language: str) -> list[str]:
    """Return a list of languages look up the translations from."""
    codes = language.replace("-", "_").split("_")
    languages = [codes[0].lower()]
    if len(codes) >= 2:
        languages.insert(0, f"{codes[0].lower()}_{codes[1].upper()}")
    if DEFAULT_LANGUAGE not in languages:
        languages.append(DEFAULT_LANGUAGE)
    return [LANGUAGE_ALIAS.get(lang, lang) for lang in languages]


def string(language: str, file: str, tid: str) -> str:
    """Translate a string identified by language, file and id."""
    if "." in tid:
        file, tid = tid.split(".", 1)
    languages = get_language_lookup(language)
    for lang_file in (file, DEFAULT_FILE):
        for lang in languages:
            source = TRANSLATIONS.get(lang, {}).get(lang_file, {})
            if tid in source:
                return source[tid]
    raise KeyError(f"The translation id '{tid}' was not to be found in any file.")


def html(language: str, file: str, tid: str, **template_replacements) -> str:
    """Translate the string identified
    by language, file and id and return an html element.

    Any id ending in -html will not be escaped but treated as raw HTML.
    """
    inner_text = string(language, file, tid)
    if template_replacements:
        inner_text = inner_text.format(**template_replacements)
    if not tid.endswith("-html"):
        inner_text = escape(inner_text)
    tid = escape(tid)
    return Markup(f'<span id="translate-{tid}" class="translation">{inner_text}</span>')


CALENDAR_LABELS = [
    "dhx_cal_today_button",
    "day_tab",
    "week_tab",
    "month_tab",
    "new_event",
    "icon_save",
    "icon_cancel",
    "icon_details",
    "icon_edit",
    "icon_delete",
    "confirm_closing",
    "confirm_deleting",
    "section_description",
    "section_time",
    "full_day",
    "confirm_recurring",
    "section_recurring",
    "button_recurring",
    "button_recurring_open",
    "button_edit_series",
    "button_edit_occurrence",
    "agenda_tab",
    "date",
    "description",
    "year_tab",
    "week_agenda_tab",
    "grid_tab",
    "drag_to_create",
    "drag_to_move",
    "message_ok",
    "message_cancel",
    "next",
    "prev",
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "repeat_radio_day",
    "repeat_radio_week",
    "repeat_radio_month",
    "repeat_radio_year",
    "repeat_radio_day_type",
    "repeat_text_day_count",
    "repeat_radio_day_type2",
    "repeat_week",
    "repeat_text_week_count",
    "repeat_radio_month_type",
    "repeat_radio_month_start",
    "repeat_text_month_day",
    "repeat_text_month_count",
    "repeat_text_month_count2_before",
    "repeat_text_month_count2_after",
    "repeat_year_label",
    "select_year_day2",
    "repeat_text_year_day",
    "select_year_month",
    "repeat_radio_end",
    "repeat_text_occurences_count",
    "repeat_radio_end2",
    "repeat_radio_end3",
]


def dhtmlx(language: str):
    """Create a dhtmlx scheduler custom locale from our translations.

    See also https://docs.dhtmlx.com/scheduler/localization.html
    """

    def cal(tid):
        """Shortcut to get an id for the calendar."""
        return string(language, CALENDAR_FILE, tid)

    result = {
        "labels": {
            "month_for_recurring": [
                cal("labels_month_for_recurring_jan"),
                cal("labels_month_for_recurring_feb"),
                cal("labels_month_for_recurring_mar"),
                cal("labels_month_for_recurring_apr"),
                cal("labels_month_for_recurring_may"),
                cal("labels_month_for_recurring_jun"),
                cal("labels_month_for_recurring_jul"),
                cal("labels_month_for_recurring_aug"),
                cal("labels_month_for_recurring_sep"),
                cal("labels_month_for_recurring_oct"),
                cal("labels_month_for_recurring_nov"),
                cal("labels_month_for_recurring_dec"),
            ],
            "day_for_recurring": [
                cal("labels_day_for_recurring_sun"),
                cal("labels_day_for_recurring_mon"),
                cal("labels_day_for_recurring_tue"),
                cal("labels_day_for_recurring_wed"),
                cal("labels_day_for_recurring_thu"),
                cal("labels_day_for_recurring_fri"),
                cal("labels_day_for_recurring_sat"),
            ],
        },
        "date": {
            "month_full": [
                cal("date_month_full_jan"),
                cal("date_month_full_feb"),
                cal("date_month_full_mar"),
                cal("date_month_full_apr"),
                cal("date_month_full_may"),
                cal("date_month_full_jun"),
                cal("date_month_full_jul"),
                cal("date_month_full_aug"),
                cal("date_month_full_sep"),
                cal("date_month_full_oct"),
                cal("date_month_full_nov"),
                cal("date_month_full_dec"),
            ],
            "month_short": [
                cal("date_month_short_jan"),
                cal("date_month_short_feb"),
                cal("date_month_short_mar"),
                cal("date_month_short_apr"),
                cal("date_month_short_may"),
                cal("date_month_short_jun"),
                cal("date_month_short_jul"),
                cal("date_month_short_aug"),
                cal("date_month_short_sep"),
                cal("date_month_short_oct"),
                cal("date_month_short_nov"),
                cal("date_month_short_dec"),
            ],
            "day_full": [
                cal("date_day_full_sun"),
                cal("date_day_full_mon"),
                cal("date_day_full_tue"),
                cal("date_day_full_wed"),
                cal("date_day_full_thu"),
                cal("date_day_full_fri"),
                cal("date_day_full_sat"),
            ],
            "day_short": [
                cal("date_day_short_sun"),
                cal("date_day_short_mon"),
                cal("date_day_short_tue"),
                cal("date_day_short_wed"),
                cal("date_day_short_thu"),
                cal("date_day_short_fri"),
                cal("date_day_short_sat"),
            ],
        },
    }
    for label in CALENDAR_LABELS:
        result["labels"][label] = cal("labels_" + label)
    return result


def dhtmlx_languages() -> list:
    """Return tuples of language name and language code."""
    result = set()
    for tid in ("language", "language-en"):
        default = string(DEFAULT_LANGUAGE, "calendar", tid)
        for code in TRANSLATIONS:
            language = string(code, "calendar", tid)
            if language != default or code == DEFAULT_LANGUAGE:
                result.add((language, code))
    result = list(result)
    result.sort()
    return result


FILES = tuple(TRANSLATIONS[DEFAULT_LANGUAGE])


def strings_translated(language, files=FILES) -> int:
    """Return the number of translations strings."""
    language = LANGUAGE_ALIAS.get(language, language)
    return sum(len(TRANSLATIONS[language].get(file, {})) for file in files)


def fraction_translated(language, files=FILES) -> float:
    """Return the 0 <= fraction <= 1 of translation."""
    return (
        1.0
        * strings_translated(language, files)
        / strings_translated(DEFAULT_LANGUAGE, files)
    )


def languages_for_the_index_file(minimal_fraction_translated=0.5):
    """Return a list of tuples of language name and code for all languages
    that are translated enough to offer the to a user.
    (language name, code, translated%)
    """
    files = ("index", "common")
    result = []
    for language, code in dhtmlx_languages():
        fraction = fraction_translated(code, files=files)
        if fraction >= minimal_fraction_translated:
            for other in result[:]:
                if other[1] == code:
                    # merge languages with duplicate code
                    language = language + "/" + other[0]  # noqa: PLW2901
                    result.remove(other)
            result.append([language, code, int(fraction * 100)])
    return result


LANGUAGE_CODES = [
    code.replace("_", "-") for code in itertools.chain(TRANSLATIONS, LANGUAGE_ALIAS)
]
LANGUAGE_CODES.sort()

__all__ = [
    "html",
    "string",
    "dhtmlx",
    "dhtmlx_languages",
    "fraction_translated",
    "strings_translated",
    "languages_for_the_index_file",
    "LANGUAGE_CODES",
]


if __name__ == "__main__":
    print("these languages are available:")  # noqa: T201
    print(LANGUAGE_CODES)  # noqa: T201
    for language in sorted(TRANSLATIONS):
        print(  # noqa: T201
            f"{language} is {int(fraction_translated(language) * 100)}% "
            f"translated: {strings_translated(language)}/"
            f"{strings_translated(DEFAULT_LANGUAGE)}"
        )
    print("These languages will be offered to the user: ")  # noqa: T201
    for lang, code, percent in languages_for_the_index_file():
        print(f"{code}\t{percent}%\t{lang}")  # noqa: T201



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""
Test the version of the Open Web Calendar.
"""

import sys

import pytest

import open_web_calendar
from open_web_calendar.app import DEFAULT_SPECIFICATION


@pytest.mark.parametrize("no_version_file", [True, False])
def test_version_can_be_imported(monkeypatch, no_version_file):
    """Import the version"""
    if no_version_file:
        monkeypatch.delitem(sys.modules, "open_web_calendar")
        monkeypatch.setitem(sys.modules, "open_web_calendar._version", None)
    import open_web_calendar

    assert open_web_calendar.version_tuple == open_web_calendar.__version_tuple__
    assert isinstance(open_web_calendar.__version_tuple__, tuple)
    assert isinstance(open_web_calendar.__version__, str)


@pytest.mark.parametrize("page", ["about.html", "index.html"])
def test_version_in_footer(page, client):
    """Check that we can see the version on the website."""
    response = client.get(f"/{page}")
    assert open_web_calendar.__version__ in response.text


@pytest.mark.parametrize(
    ("version_specified", "version_seen"),
    [
        ("", open_web_calendar.__version__),
        ("v1000.1000BETA", "v1000.1000BETA"),
    ],
)
@pytest.mark.parametrize("page", ["about.html", "index.html"])
def test_version_from_default_specification_is_used(
    page, version_seen, version_specified, client, monkeypatch
):
    """The version should also be possible to specify in the default spec."""
    monkeypatch.setitem(DEFAULT_SPECIFICATION, "version", version_specified)
    response = client.get(f"/{page}")
    assert version_seen in response.text



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Test adding JavaScript to the calendar.

This allows people to
- add DHTMLX scheduler customizations
- configure the scheduler themselves
"""

from urllib.parse import quote

import pytest


@pytest.mark.parametrize(
    ("query", "tags"),
    [
        (
            "?javascript_url=https://tippi.js/embed.js",
            ['<script src="https://tippi.js/embed.js"'],
        ),
        (
            "?javascript_url=/feature1.js&javascript_url=/feature2.js",
            ['<script src="/feature1.js"', '<script src="/feature2.js"'],
        ),
    ],
)
def test_embed_js_link(client, query, tags):
    """Check that the JS links are used."""
    response = client.get(f"/calendar.html{query}")
    for html in tags:
        assert html in response.text


@pytest.mark.parametrize(
    ("js", "html"),
    [
        ("console.log(specification);", "console.log(specification);"),
        ("<html>&", "&lt;html&gt;&amp;"),
    ],
)
def test_embed_js_directly(client, js, html):
    """Check direct JS embedding."""
    response = client.get(f"/calendar.html?javascript={quote(js)}")
    print(response.text)
    assert f'<script type="text/javascript">{html}</script>' in response.text


@pytest.mark.parametrize(
    ("query", "urls"),
    [
        ("?css_url=https://tippi.js/embed.css", ["https://tippi.js/embed.css"]),
        (
            "?css_url=/feature1.css&css_url=/feature2.css",
            ["/feature1.css", "/feature2.css"],
        ),
    ],
)
def test_embed_css_link(client, query, urls):
    """Check that the CSS links are used."""
    response = client.get(f"/calendar.html{query}")
    for url in urls:
        assert (
            f'<link href="{url}" rel="stylesheet" type="text/css" charset="utf-8">'
            in response.text
        )



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""
Test the caching functionality which will be used by subsequent tests.

"""

from typing import NamedTuple

import pytest
import requests

from open_web_calendar.app import cache_url, get_text_from_url


# use the content property
# see https://stackoverflow.com/a/18810889/1320237
class MockRequestResult(NamedTuple):
    content: str


def test_requests_are_automatically_cached(monkeypatch, mock):
    mock.return_value = MockRequestResult("trallala")
    monkeypatch.setattr(requests, "get", mock)
    assert get_text_from_url("https://asd.asd") == "trallala"
    assert get_text_from_url("https://asd.asd") == "trallala"
    mock.assert_called_once()


def test_tests_can_cache_values():
    cache_url("http://asd.asd", "test123")
    assert get_text_from_url("http://asd.asd") == "test123"


def test_caching_twice_works():
    cache_url("http://asd.asd", "test123")
    cache_url("http://asd.asd", "asd")
    assert get_text_from_url("http://asd.asd") == "asd"


def test_requests_are_forbidden():
    import requests

    with pytest.raises(RuntimeError):
        requests.get("https://duckduckgo.com", timeout=20)



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""Test that special time zones are also handeled correctly.

See https://github.com/niccokunzmann/open-web-calendar/issues/190
"""

import pytest


@pytest.mark.parametrize(
    ("timezone", "event_start", "timeshift"),
    [
        ("", "2019-03-04 07:00", 0),  # UTC
        ("", "2019-03-04 04:00", 180),  # UTC-3
        ("Etc/GMT 3", "2019-03-04 04:00", 180),
        ("Etc/GMT 5", "2019-03-04 02:00", 300),
        ("undefined", "2019-03-04 06:00", 60),
        ("undefined", "2019-03-04 07:00", 0),
        ("UTC", "2019-03-04 07:00", 180),  # timeshift ignored, proper timezone
        (
            "Europe/London",
            "2019-03-04 07:00",
            -240,
        ),  # timeshift ignored, proper timezone
        ("Europe/Berlin", "2019-03-04 08:00", 0),  # timeshift ignored, proper timezone
    ],
)
def test_that_timezones_do_not_create_an_error(
    timezone, client, event_start, calendar_urls, timeshift
):
    """These time zones are a bit special, not 1:1 Python.

    This is why we use the timeshift given.
    """
    response = client.get(
        f"/calendar.events.json?url={calendar_urls['one-event']}&timezone={timezone}&timeshift={timeshift}&from=2019-03-04&to=2019-03-05"
    )
    assert response.status_code == 200
    assert len(response.json) == 1
    event = response.json[0]
    assert event["start_date"] == event_start, f"Timezone differs in offset: {timezone}"



# SPDX-FileCopyrightText: 2024 Nicco Kunzmann and Open Web Calendar Contributors <https://open-web-calendar.quelltext.eu/>
#
# SPDX-License-Identifier: GPL-2.0-only

"""We want to be able to subscribe from sites that provide ICS links indirectly.

See https://github.com/niccokunzmann/open-web-calendar/issues/309
"""


def test_request_from_alternate_link(client, calendar_urls):
    """If we request from an alternate link, we should see the events from the calendar.

    The HTML page links to the ics file.
    """
    calendar_text = client.get(
        f"/calendar.ics?url={calendar_urls['one-day-event']}"
    ).text
    html_text = client.get(
        f"/calendar.ics?url={calendar_urls['one-day-event.html']}"
    ).text
    assert (
        calendar_text == html_text
    ), "Whether alternate link or direct, the content should be the same."
